@@ -10,7 +10,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tclRegexp.c,v 1.2.2.2 1999/05/20 02:38:27 redman Exp $
+ * RCS: @(#) $Id: tclRegexp.c,v 1.9 1999/06/17 19:32:15 stanton Exp $
  */
 
 #include "tclInt.h"
@@ -94,6 +94,9 @@ static void		DupRegexpInternalRep _ANSI_ARGS_((Tcl_Obj *srcPtr,
 static void		FinalizeRegexp _ANSI_ARGS_((ClientData clientData));
 static void		FreeRegexp _ANSI_ARGS_((TclRegexp *regexpPtr));
 static void		FreeRegexpInternalRep _ANSI_ARGS_((Tcl_Obj *objPtr));
+static int		RegExpExecUniChar _ANSI_ARGS_((Tcl_Interp *interp,
+			    Tcl_RegExp re, CONST Tcl_UniChar *uniString,
+			    int numChars, int nmatches, int flags));
 static int		SetRegexpFromAny _ANSI_ARGS_((Tcl_Interp *interp,
 			    Tcl_Obj *objPtr));
 
@@ -177,31 +180,39 @@ Tcl_RegExpExec(interp, re, string, start)
 				 * this identifies beginning of larger
 				 * string, so that "^" won't match. */
 {
-    int result, numChars;
-    Tcl_DString stringBuffer;
-    Tcl_UniChar *uniString;
-
-    TclRegexp *regexpPtr = (TclRegexp *) re;
+    int flags, result, numChars;
+    TclRegexp *regexp = (TclRegexp *)re;
+    Tcl_DString ds;
+    Tcl_UniChar *ustr;
 
     /*
-     * Remember the UTF-8 string so Tcl_RegExpRange() can convert the
-     * matches from character to byte offsets.
+     * If the starting point is offset from the beginning of the buffer,
+     * then we need to tell the regexp engine not to match "^".
      */
 
-    regexpPtr->string = string;
-
-    Tcl_DStringInit(&stringBuffer);
-    uniString = Tcl_UtfToUniCharDString(string, -1, &stringBuffer);
-    numChars = Tcl_DStringLength(&stringBuffer) / sizeof(Tcl_UniChar);
+    if (string > start) {
+	flags = REG_NOTBOL;
+    } else {
+	flags = 0;
+    }
 
     /*
-     * Perform the regexp match.
+     * Remember the string for use by Tcl_RegExpRange().
      */
 
-    result = TclRegExpExecUniChar(interp, re, uniString, numChars, -1,
-	    ((string > start) ? REG_NOTBOL : 0));
+    regexp->string = string;
+    regexp->objPtr = NULL;
 
-    Tcl_DStringFree(&stringBuffer);
+    /*
+     * Convert the string to Unicode and perform the match.
+     */
+
+    Tcl_DStringInit(&ds);
+    ustr = Tcl_UtfToUniCharDString(string, -1, &ds);
+    numChars = Tcl_DStringLength(&ds) / sizeof(Tcl_UniChar);
+    result = RegExpExecUniChar(interp, re, ustr, numChars,
+	    -1 /* nmatches */, flags);
+    Tcl_DStringFree(&ds);
 
     return result;
 }
@@ -238,23 +249,27 @@ Tcl_RegExpRange(re, index, startPtr, endPtr)
 				 * in (sub-) range here. */
 {
     TclRegexp *regexpPtr = (TclRegexp *) re;
+    CONST char *string;
 
     if ((size_t) index > regexpPtr->re.re_nsub) {
 	*startPtr = *endPtr = NULL;
     } else if (regexpPtr->matches[index].rm_so < 0) {
 	*startPtr = *endPtr = NULL;
     } else {
-	*startPtr = Tcl_UtfAtIndex(regexpPtr->string,
-		regexpPtr->matches[index].rm_so);
-	*endPtr = Tcl_UtfAtIndex(regexpPtr->string,
-		regexpPtr->matches[index].rm_eo);
+	if (regexpPtr->objPtr) {
+	    string = Tcl_GetString(regexpPtr->objPtr);
+	} else {
+	    string = regexpPtr->string;
+	}
+	*startPtr = Tcl_UtfAtIndex(string, regexpPtr->matches[index].rm_so);
+	*endPtr = Tcl_UtfAtIndex(string, regexpPtr->matches[index].rm_eo);
     }
 }
 
 /*
  *---------------------------------------------------------------------------
  *
- * TclRegExpExecUniChar --
+ * RegExpExecUniChar --
  *
  *	Execute the regular expression matcher using a compiled form of a
  *	regular expression and save information about any match that is
@@ -272,8 +287,8 @@ Tcl_RegExpRange(re, index, startPtr, endPtr)
  *----------------------------------------------------------------------
  */
 
-int
-TclRegExpExecUniChar(interp, re, wString, numChars, nmatches, flags)
+static int
+RegExpExecUniChar(interp, re, wString, numChars, nmatches, flags)
     Tcl_Interp *interp;		/* Interpreter to use for error reporting. */
     Tcl_RegExp re;		/* Compiled regular expression; returned by
 				 * a previous call to Tcl_GetRegExpFromObj */
@@ -287,13 +302,15 @@ TclRegExpExecUniChar(interp, re, wString, numChars, nmatches, flags)
 {
     int status;
     TclRegexp *regexpPtr = (TclRegexp *) re;
-    size_t nm = regexpPtr->re.re_nsub + 1;
+    size_t last = regexpPtr->re.re_nsub + 1;
+    size_t nm = last;
 
-    if (nmatches >= 0 && (size_t) nmatches < nm)
+    if (nmatches >= 0 && (size_t) nmatches < nm) {
 	nm = (size_t) nmatches;
+    }
 
     status = TclReExec(&regexpPtr->re, wString, (size_t) numChars,
-	    (rm_detail_t *)NULL, nm, regexpPtr->matches, flags);
+	    &regexpPtr->details, nm, regexpPtr->matches, flags);
 
     /*
      * Check for errors.
@@ -318,12 +335,13 @@ TclRegExpExecUniChar(interp, re, wString, numChars, nmatches, flags)
  * TclRegExpRangeUniChar --
  *
  *	Returns pointers describing the range of a regular expression match,
- *	or one of the subranges within the match.
+ *	or one of the subranges within the match, or the hypothetical range
+ *	represented by the rm_extend field of the rm_detail_t.
  *
  * Results:
  *	The variables at *startPtr and *endPtr are modified to hold the
- *	addresses of the endpoints of the range given by index.  If the
- *	specified range doesn't exist then NULLs are returned.
+ *	offsets of the endpoints of the range given by index.  If the
+ *	specified range doesn't exist then -1s are supplied.
  *
  * Side effects:
  *	None.
@@ -337,7 +355,8 @@ TclRegExpRangeUniChar(re, index, startPtr, endPtr)
 				 * been passed to Tcl_RegExpExec. */
     int index;			/* 0 means give the range of the entire
 				 * match, > 0 means give the range of
-				 * a matching subrange. */
+				 * a matching subrange, -1 means the
+				 * range of the rm_extend field. */
     int *startPtr;		/* Store address of first character in
 				 * (sub-) range here. */
     int *endPtr;		/* Store address of character just after last
@@ -345,7 +364,10 @@ TclRegExpRangeUniChar(re, index, startPtr, endPtr)
 {
     TclRegexp *regexpPtr = (TclRegexp *) re;
 
-    if ((size_t) index > regexpPtr->re.re_nsub) {
+    if ((regexpPtr->flags&REG_EXPECT) && index == -1) {
+	*startPtr = regexpPtr->details.rm_extend.rm_so;
+	*endPtr = regexpPtr->details.rm_extend.rm_eo;
+    } else if ((size_t) index > regexpPtr->re.re_nsub) {
 	*startPtr = -1;
 	*endPtr = -1;
     } else {
@@ -375,7 +397,7 @@ TclRegExpRangeUniChar(re, index, startPtr, endPtr)
 
 int
 Tcl_RegExpMatch(interp, string, pattern)
-    Tcl_Interp *interp;		/* Used for error reporting. */
+    Tcl_Interp *interp;		/* Used for error reporting. May be NULL. */
     char *string;		/* String. */
     char *pattern;		/* Regular expression to match against
 				 * string. */
@@ -392,9 +414,9 @@ Tcl_RegExpMatch(interp, string, pattern)
 /*
  *----------------------------------------------------------------------
  *
- * TclRegExpMatchObj --
+ * Tcl_RegExpExecObj --
  *
- *	See if a string matches a regular expression pattern object.
+ *	Execute a precompiled regexp against the given object.
  *
  * Results:
  *	If an error occurs during the matching operation then -1
@@ -403,25 +425,111 @@ Tcl_RegExpMatch(interp, string, pattern)
  *	and 0 otherwise.
  *
  * Side effects:
- *	None.
+ *	Converts the object to a Unicode object.
  *
  *----------------------------------------------------------------------
  */
 
 int
-TclRegExpMatchObj(interp, string, patObj)
-    Tcl_Interp *interp;		/* Used for error reporting. */
-    char *string;		/* String. */
-    Tcl_Obj *patObj;		/* Regular expression to match against
+Tcl_RegExpExecObj(interp, re, objPtr, offset, nmatches, flags)
+    Tcl_Interp *interp;		/* Interpreter to use for error reporting. */
+    Tcl_RegExp re;		/* Compiled regular expression;  must have
+				 * been returned by previous call to
+				 * Tcl_GetRegExpFromObj. */
+    Tcl_Obj *objPtr;		/* String against which to match re. */
+    int offset;			/* Character index that marks where matching
+				 * should begin. */
+    int nmatches;		/* How many subexpression matches (counting
+				 * the whole match as subexpression 0) are
+				 * of interest.  -1 means all of them. */
+    int flags;			/* Regular expression execution flags. */
+{
+    TclRegexp *regexpPtr = (TclRegexp *) re;
+    Tcl_UniChar *udata;
+    int length;
+
+    /*
+     * Save the target object so we can extract strings from it later.
+     */
+
+    regexpPtr->string = NULL;
+    regexpPtr->objPtr = objPtr;
+
+    udata = Tcl_GetUnicode(objPtr);
+    length = Tcl_GetCharLength(objPtr);
+
+    if (offset > length) {
+	offset = length;
+    }
+    udata += offset;
+    length -= offset;
+    
+    return RegExpExecUniChar(interp, re, udata, length, nmatches, flags);
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Tcl_RegExpMatchObj --
+ *
+ *	See if an object matches a regular expression.
+ *
+ * Results:
+ *	If an error occurs during the matching operation then -1
+ *	is returned and the interp's result contains an error message.
+ *	Otherwise the return value is 1 if "string" matches "pattern"
+ *	and 0 otherwise.
+ *
+ * Side effects:
+ *	Changes the internal rep of the pattern and string objects.
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+Tcl_RegExpMatchObj(interp, stringObj, patternObj)
+    Tcl_Interp *interp;		/* Used for error reporting. May be NULL. */
+    Tcl_Obj *stringObj;		/* Object containing the String to search. */
+    Tcl_Obj *patternObj;	/* Regular expression to match against
 				 * string. */
 {
     Tcl_RegExp re;
 
-    re = Tcl_GetRegExpFromObj(interp, patObj, REG_ADVANCED);
+    re = Tcl_GetRegExpFromObj(interp, patternObj,
+	    TCL_REG_ADVANCED | TCL_REG_NOSUB);
     if (re == NULL) {
 	return -1;
     }
-    return Tcl_RegExpExec(interp, re, string, string);
+    return Tcl_RegExpExecObj(interp, re, stringObj, 0 /* offset */,
+	    0 /* nmatches */, 0 /* flags */);
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Tcl_RegExpGetInfo --
+ *
+ *	Retrieve information about the current match.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+Tcl_RegExpGetInfo(regexp, infoPtr)
+    Tcl_RegExp regexp;		/* Pattern from which to get subexpressions. */
+    Tcl_RegExpInfo *infoPtr;	/* Match information is stored here.  */
+{
+    TclRegexp *regexpPtr = (TclRegexp *) regexp;
+
+    infoPtr->nsubs = regexpPtr->re.re_nsub;
+    infoPtr->matches = (Tcl_RegExpIndices *) regexpPtr->matches;
+    infoPtr->extendStart = regexpPtr->details.rm_extend.rm_so;
 }
 
 /*
@@ -784,6 +892,10 @@ CompileRegexp(interp, string, length, flags)
      */
     
     regexpPtr = (TclRegexp *) ckalloc(sizeof(TclRegexp));
+    regexpPtr->objPtr = NULL;
+    regexpPtr->string = NULL;
+    regexpPtr->details.rm_extend.rm_so = -1;
+    regexpPtr->details.rm_extend.rm_eo = -1;
 
     /*
      * Get the up-to-date string representation and map to unicode.
